#include "huffman.h"
#include <stdio.h>

int huffman(unsigned char * inbuf, unsigned char * outbuf, unsigned int size, code CODES[256]) {
    int i, j, n;
    node TREE[257][256];
    for (i=0; i<256; i++) {
	TREE[0][i].value = 0; //clear first row
	TREE[0][i].bit = -1;
    }
    count_chars(inbuf, TREE[0], size); //count number of occurences
    j = 0;
    for (i=0; i<256; i++) { //create row of non-zero entries
	if (TREE[0][i].value!=0) {
	    TREE[0][i].parent = &(TREE[1][j]);
	    TREE[1][j].value = TREE[0][i].value; //copy first row to the secend one
	    j++;
	}
    }
    create_tree(TREE, j); // create Huffman tree
    get_codes(TREE, CODES); //get bit-codes from the Huffman tree
    n = put_bits(inbuf, outbuf, CODES, size);
    return n;
}

int count_chars(unsigned char * buf, node * amounts, unsigned int size) {
    int i ,j;
    for (i=0; i<size; i++) { //simple code counting characters
	j=*(buf+i);
	(amounts + j)->value++;
    }
    return 0;
}

int create_tree(node TREE[256][256], int size) {
    int i, j, k; 
    int a, b;
    for (i=1; i<(size-1); i++) {
	lowest(TREE[i], &a, &b, size-i+1); //find lowest values
	if (b<a) { //we need b>a so we use a nice trick to replace them
	    b = b+a;
	    a = b-a;
	    b = b-a;
	}
        TREE[i+1][a].value = TREE[i][a].value + TREE[i][b].value; //write sum to next node
        TREE[i][a].parent = &(TREE[i+1][a]); //set parent of left item
        TREE[i][b].parent = &(TREE[i+1][a]); //set parent of right item
	if (TREE[i][a].value < TREE[i][b].value) { //mark left and right leave
	    TREE[i][a].bit = 0;
	    TREE[i][b].bit = 1;
	} else {
	    TREE[i][a].bit = 1;
	    TREE[i][b].bit = 0;
	}
	j = 0;
	for (k=0; k<=(size-i); k++) { //unchanged values
	    if (k==b) {
		j = 1; //since now there is a shift in copying
	    } else {
		if (k!=a) {	
		    TREE[i][k].parent = &(TREE[i+1][k-j]); //set parent
		    TREE[i+1][k-j].value = TREE[i][k].value; //copy value
		    TREE[i][k].bit = -1;
		}
	    }
	}
    }
    if (TREE[size-1][0].value < TREE[size-1][1].value) { //we set properties of last row
	TREE[size-1][0].bit = 0;
	TREE[size-1][1].bit = 1;
    } else {
	TREE[size-1][0].bit = 1;
	TREE[size-1][1].bit = 0;
    }
    TREE[size-1][0].parent = NULL;
    TREE[size-1][1].parent = NULL;
    
}

int lowest(node TAB[256], int * a, int * b, int size) { //get indexes of two lowest values
    unsigned int c ,d;
    int i;
    c = d = 0xFFFFFFFF;
    *a = *b = 0;
    for (i=0; i<size; i++) {
	if (TAB[i].value<c) {
	    d = c;
	    c = TAB[i].value;
	    *b = *a;
	    *a = i;
	} else {
	    if (TAB[i].value<d) {
		d = TAB[i].value;
		*b = i;
	    }
	}
    }
    return 0;
}

int get_codes(node TREE[256][256], code CODES[256]) { //get bit-codes for each byte using Huffman tree
    node * p_node;
    int i;
    for (i=0; i<256; i++) {
	p_node = &(TREE[0][i]);
	CODES[i].mask=0;
	while (p_node!=NULL) {
	    if (p_node->bit==0) {
		CODES[i].value = CODES[i].value<<1;
		CODES[i].value = CODES[i].value & 0xFFFFFFFE; //put one bit to the mask
		CODES[i].mask++; //increment number of bits in the mask
	    }
	    if (p_node->bit==1) {
		CODES[i].value = CODES[i].value<<1;
		CODES[i].value = CODES[i].value | 0x00000001;
		CODES[i].mask++;
	    }
	    p_node = p_node->parent; //go to next node
	}
    }
    return 0;
}

int put_bits(unsigned char * inbuf, unsigned char * outbuf,  code * bitcode, unsigned int size) {
    
    unsigned int i, j, k;
    i = 0;
    k = 0;
    int n = 0; //number of compressed bytes
    unsigned char temp; //actual byte
    unsigned int temp2; //bit-code
    
    j = (bitcode+(*(inbuf)))->mask; //get amount of bits in current bit-code
    temp2 = (bitcode+(*(inbuf)))->value; //take first bitcode
    while (k < size) {
	temp = temp | temp2 & 0x00000001;
	temp2 = temp2 >> 1;
	j--;
	i++;
	if ((i%8)==0) {;
	    *outbuf = temp;
	    outbuf++;
	    i = 0;
	    n++;
	}
	temp = temp << 1;
	if (j==0) {
	    inbuf++;
	    k++;
	    j = (bitcode+(*(inbuf)))->mask; //get amount of bits in current bit-code
	    temp2 = (bitcode+(*(inbuf)))->value; //take next bitcode
	}
    }
    return n;
}

int get_bits(unsigned char * inbuf, unsigned char * outbuf, code * bitcode, unsigned int size) {

    unsigned char temp; //actual byte
    unsigned int temp2 = 0; //bit-code
    int i;
    int j = 0; //read bits, total
    int k = 0; //read bits in current byte
    int n = 0; //decoded characters

    while (n<size) {
	if (k==0) {
	    temp = *inbuf;
	    inbuf++;
	}
        if (temp & 0x80) temp2 = temp2 | 0x00000001; else temp2 = temp2 & 0xFFFFFFFE; //get one bit from byte
        k++;
        j++;
        k = k % 8;
        for (i=0; i<256; i++) {
    	    if (j==(bitcode+i)->mask) { //look for code of the same length
    		if ((temp2 | (0xFFFFFFFF << (bitcode+i)->mask)) == (bit_reverse((bitcode+i)->value, (bitcode+i)->mask) | (0xFFFFFFFF << (bitcode+i)->mask))) {
		    //we have found match
		    *outbuf = i;
		    outbuf++;
		    n++;
		    j = 0;
		    i = 256;
		}
	    }
	}
	temp2 <<= 1; //go to next bit
        temp <<= 1;
    }
    return n;
}

unsigned int bit_reverse(unsigned int a, int size) { //reverse order of bits
    unsigned int o;
    int i;
    for (i=0; i<size; i++) {
	o <<= 1;
	if (a & 0x00000001) o = o | 0x00000001; else o = o & 0xFFFFFFFE;
	a >>= 1;
    }
    return o;
}
